# Generated by Django 5.1.2 on 2024-10-10 09:24

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("codenerix_email", "0011_alter_emailmessage_content_subtype_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="emailmessage",
            name="unsubscribe_url",
            field=models.URLField(
                blank=True, null=True, verbose_name="Unsubscribe URL"
            ),
        ),
    ]
